(* Turbo51-Unit
  interrupt buffered reading from serial port (standard in)
  optimized version using assembler parts

  Important:  Main program must contain directive "{$DefaultFile On }"
  ========== 
              requires extended memory

  J. Rathlev, Jan. 2011
  *)

unit SerialInOut;

{$XDATA}

// ----------------------------------------------------------------- 
interface 

{$I CpuClock.inc}

const
  DefBaudRate = 9600;
  BufLen = 256;   // length of input buffer
  Lf = 10;
  Cr = 13;

function KeyPressed : boolean;
function ReadKey : char; 

procedure SetBaudRate (Baudrate : longint);
procedure WriteToSerialPort;
function ReadFromSerialPort: Char;
function ReceivedChars : byte;
function ReceivedLines: byte;

procedure SerialInt; interrupt Serial; using 3;

// ----------------------------------------------------------------- 
implementation

type
  TBuffer = array [0..BufLen-1] of byte;

var
  BaudRateTimerValue : Byte;
  ICCnt,ILCnt : word; volatile;
  IWPnt,IRPnt : Pointer;
  InBuf : TBuffer  XDATA; volatile;
  OutRdy : boolean; volatile;
  
// ----------------------------------------------------------------- 
// interrupt procedure receive character from serial port
procedure SerialInt; Assembler;
Asm
    PUSH      PSW
    MOV       PSW, #$18
    PUSH      ACC
    PUSH      DPL
    PUSH      DPH
// if RI then begin    
    JNB       RI, @SiSend
    CLR       RI
    MOV       R2, SBUF
//  if c=Lf then Exit;             // ignore line feed  
    MOV       A, R2
    XRL       A, #Lf  
    JZ        @SiSend
// if ICCnt>=BufLen then Exit;    // buffer overflow - ignore char
    MOV       A, ICCnt
    ADD       A, #LOW (-BufLen)
    MOV       A, ICCnt+1
    ADDC      A, #HIGH (-BufLen)
    JC        @SiSend
// if c=CR then inc(ILCnt);       // count lines
    CJNE      R2, #Cr, @SiNxt2
    INC       ILCnt
    MOV       A, ILCnt
    JNZ       @SiNxt2
    INC       ILCnt+1
// InBuf[IWPnt]:=c;    
@SiNxt2:
    MOV       DPL, IWPnt
    MOV       DPH, IWPnt+1
    MOV       A, R2
    MOVX      @DPTR, A
// inc(IWPnt);
    INC       DPTR
// if IWPnt=InBuf+BufLen then IWPnt:=InBuf
@SiNxt3:
    MOV       A, DPH
    XRL       A, #HIGH(Inbuf+BufLen)
    JNZ       @SiNxt4
    MOV       A, DPL
    XRL       A, #LOW(Inbuf+BufLen)
    JNZ       @SiNxt4
    MOV       DPL, #LOW(InBuf)
    MOV       DPH, #HIGH(InBuf)   
@SiNxt4:    
    MOV       IWPNT, DPL
    MOV       IWPNT+1, DPH
// inc(ICCnt);
    INC       ICCnt
    MOV       A, ICCnt
    JNZ       @SiSend
    INC       ICCnt+1
@SiSend:
    JNB       TI, @SiRet
    CLR       TI 
    SETB      OutRdy
@SiRet:
    POP       DPH
    POP       DPL
    POP       ACC
    POP       PSW
    RETI
  end;
  
// ----------------------------------------------------------------- 
// Basic I/O routines for "Assign"
// send character to serial port (polling)
procedure WriteToSerialPort; Assembler;
Asm
@WaitLoop:
    JNB       OutRdy, @WaitLoop  // wait for TI cleared by interrupt
    MOV       SBUF, A
    CLR       OutRdy
  end;

// receive character from  serial port (via interrupt)
function ReadFromSerialPort: Char; Assembler;
Asm
    PUSH      DPL
    PUSH      DPH
    PUSH      AR2
// repeat until ICCnt<>0;
@RsLoop:
    MOV       A, ICCnt
    JNZ       @RsNxt2
    MOV       A, ICCnt+1
    JZ        @RsLoop   
// dec(ICCnt);
@RsNxt2:
    DEC       ICCnt
    MOV       A, ICCnt
    CJNE      A, #-1, @RsNxt1
    DEC       ICCnt+1
// ByteResult:=InBuf[IRPnt];
@RsNxt1:  
    MOV       DPL, IRPnt
    MOV       DPH, IRPnt+1
    MOVX      A, @DPTR
    MOV       R2, A
// inc(IRPnt);
    INC       DPTR
// if IWPnt=InBuf+BufLen then IWPnt:=InBuf
    MOV       A, DPH
    XRL       A, #HIGH(Inbuf+BufLen)
    JNZ       @RsNxt3
    MOV       A, DPL
    XRL       A, #LOW(Inbuf+BufLen)
    JNZ       @RsNxt3
    MOV       DPL, #LOW(InBuf)
    MOV       DPH, #HIGH(InBuf)   
// if IRPnt>=BufLen then IRPnt:=0;
@RsNxt3:
    MOV       IRPNT, DPL
    MOV       IRPNT+1, DPH
// if ByteResult=Cr then begin
    CJNE      R2, #Cr, @RsNxt4
// if ILCnt<>0 then dec(ILCnt);
    MOV       A, ILCnt
    JNZ       @RsNxt6
    MOV       A, ILCnt+1
    JZ        @RsNxt5
@RsNxt6:    
    DEC       ILCnt
    MOV       A, ILCnt
    CJNE      A, #-1, @RsNxt5
    DEC       ILCnt+1
// ByteResult:=Lf;
@RsNxt5:
    MOV       R2, #Lf
@RsNxt4:
    MOV       A, R2
    POP       AR2
    POP       DPH
    POP       DPL
  end;

// ----------------------------------------------------------------- 
// number of characters in input buffer
function ReceivedChars : byte;
begin
  Result:=ICCnt;
  end;

// number of lines in input buffer
function ReceivedLines: byte;
begin
  Result:=ILCnt;
  end;

// test if at least one character in input buffer
function KeyPressed : boolean;
begin
  Result:=ICCnt>0;
  end;

// read one character from input buffer
function ReadKey : char; 
begin
  Result:=ReadFromSerialPort;
  end;
    
// set baud rate for serial port
procedure SetBaudRate (Baudrate : longint);
begin
  BaudRateTimerValue:=256-CpuClock div (12*16*Baudrate);
  TL1  := BaudRateTimerValue;
  TH1  := BaudRateTimerValue;
  end;
  
// initialize serial port    
procedure InitSerial;
begin
  SetBaudrate(defBaudRate);
  TMOD := TMOD and $F or %00100000;    { Timer1: no GATE, 8 bit timer, autoreload }
  SCON := %01010000;    { Serial Mode 1, Enable Reception }
  PCON := PCON or $80;
  TR1  := True;         { Enable timer 1 }
  RI   := False;
  ICCnt:=0; ILCnt:=0; 
  TI   := False;
  OutRdy:=True;
  Asm
    MOV       IWPnt, #LOW(InBuf)
    MOV       IWPnt+1, #HIGH(InBuf)   
    MOV       IRPnt, #LOW(InBuf)
    MOV       IRPnt+1, #HIGH(InBuf)   
    end;
  ES:=true;  {Enable serial interrupt}
  end;

begin
  InitSerial;
  Assign (CurrentIO, ReadFromSerialPort, WriteToSerialPort);
end.